---@type Mq
mq = require('mq')
require "MqUtils"

-- Edit these two numbers if you want other choices
-- Which spell gem to use, it puts the original spell back after its done.
local GEM_SLOT = 7
-- Recast the DS even if target already has it, but has less than X percent of duration time remaining.
local RECAST_UNDER_PERCENT = 90

-- Thanks to RGMercs for the current list of spells!
local coat_names = {
    ----- 125 Mage DS highest level to lowest-----
    "Circle of Forgefire Coat",
    "Forgefire Coat",
    "Circle of Emberweave Coat",
    "Emberweave Coat",
    "Circle of Igneous Skin",
    "Igneous Coat",
    "Circle of the Inferno",
    "Inferno Coat",
    "Circle of Flameweaving",
    "Flameweave Coat",
    "Circle of Flameskin",
    "Flameskin",
    "Circle of Embers",
    "Embercoat",
    "Circle of Dreamfire",
    "Dreamfire Coat",
    "Circle of Brimstoneskin",
    "Brimstoneskin",
    "Circle of Lavaskin",
    "Lavaskin",
    "Circle of Magmaskin",
    "Magmaskin",
    "Circle of Fireskin",
    "Fireskin",
    "Maelstrom of Ro",
    "FlameShield of Ro",
    "Aegis of Ro",
    "Cadeau of Flame",
    "Boon of Immolation",
    "Shield of Lava",
    "Barrier of Combustion",
    "Inferno Shield",
    "Shield of Flame",
    "Shield of Fire",
    ----- 125 Druid DS highest level to lowest-----
    "Legacy of Bloodspikes",
    "Duskthorn",
    "Legacy of Icebriars",
    "Icebriar Bulwark",
    "Legacy of Daggerspikes",
    "Daggerspike Bulwark",
    "Legacy of Daggerspurs",
    "Daggerspur Bulwark",
    "Legacy of Spikethistles",
    "Spikethistle Bulwark",
    "Legacy of Spineburrs",
    "Spineburr Bulwark",
    "Legacy of Bonebriar",
    "Bonebriar Bulwark",
    "Legacy of Brierbloom",
    "Brierbloom Bulwark",
    "Legacy of Viridithorns",
    "Viridifloral Bulwark",
    "Legacy of Viridiflora",
    "Viridifloral Shield",
    "Legacy of Nettles",
    "Nettle Shield",
    "Legacy of Bracken",
    "Shield of Bracken",
    "Legacy of Thorn",
    "Shield of Blades",
    "Legacy of Spike",
    "Shield of Thorns",
    "Shield of Spikes",
    "Shield of Brambles",
    "Shield of Barbs",
    "Shield of Thistles",
}

local function ticksToTimeString(ticks)
    local totalSeconds = ticks / 1000
    local hours = math.floor(totalSeconds / 3600)
    local minutes = math.floor((totalSeconds % 3600) / 60)
    local seconds = totalSeconds % 60

    return string.format("%01d:%02d:%02d", hours, minutes, seconds)
end

function SpellReadyToCast(slot)
    if mq.TLO.Me.SpellInCooldown() then return false end
    if mq.TLO.Me.Casting() then return false end
    if not mq.TLO.Me.SpellReady(slot)() then return false end
    if mq.TLO.Window('SpellBookWnd').Open() then return false end
    local ticks = mq.TLO.Me.GemTimer(slot)()
    return (ticks == 0)
end

function BestDS()
    -- BestDS returns true if a spell was cast
    -- Target Check
    if mq.TLO.Target() == nil then
        print('No target for BestDS.')
        return false
    end
    -- Remember the original spell that was in the gem slot
    local original_gem = mq.TLO.Me.Gem(GEM_SLOT)()
    for _, coat_name in ipairs(coat_names) do
        coat_name = mq.TLO.Spell(coat_name).RankName()
        -- Do you know the spell and will it stack on your target.
        if mq.TLO.Me.Book(coat_name)() ~= nil and mq.TLO.Spell(coat_name).StacksTarget() then
            -- Duration Check
            local my_duration_ticks = mq.TLO.Spell(coat_name).MyDuration() * 6000
            local target_duration = mq.TLO.Target.BuffDuration(coat_name)()
            if target_duration ~= nil and (my_duration_ticks * (RECAST_UNDER_PERCENT / 100)) < target_duration then
                print(mq.TLO.Target() .. ' has recent BestDS: ' .. coat_name .. ' ' .. ticksToTimeString(target_duration))
                return false
            end
            -- Range Check
            local range = math.max(mq.TLO.Spell(coat_name).MyRange(), mq.TLO.Spell(coat_name).AERange())
            if mq.TLO.Target.Distance3D() > range then
                print(mq.TLO.Target() .. ' is out of ' .. coat_name .. ' range ' .. range .. ': '.. string.format("%.2f", mq.TLO.Target.Distance3D()))
                return false
            end
            -- Cast the DS
            print('Casting BestDS: ' .. coat_name .. ' on ' .. mq.TLO.Target() .. ' (Lvl: ' .. mq.TLO.Target.Level() .. ')')
            mq.cmd('/casting "' .. coat_name .. '" gem' .. GEM_SLOT)
            -- If we replaced GEM_SLOT with a different spell, put back the original.
            if original_gem ~= nil and original_gem ~= coat_name then
                mq.delay(1000)
                while not SpellReadyToCast(GEM_SLOT) do
                    mq.delay(100)
                end
                print('Restoring ' .. original_gem .. ' to Gem ' .. GEM_SLOT)
                mq.cmd('/memorize "' .. original_gem .. '" ' .. GEM_SLOT)
            end
            return true
        end
    end
    print('No stackable BestDS found.')
    return false
end

BestDS()
